from datetime import datetime

import pandas as pd
from dateutil import parser

from sapextractor.utils import constants
from sapextractor.utils.dates import timestamp_column_from_dt_tm
from sapextractor.utils.vbtyp import extract_vbtyp


def extract_vbak(con, min_extr_date="2020-01-01 00:00:00"):
    vbak = {}
    try:
        vbak = con.prepare_and_execute_query("VBAK", ["VBELN", "ERDAT", "ERZET"])
        timestamp_column_from_dt_tm.apply(vbak, "ERDAT", "ERZET", "event_timestamp")
        vbak = vbak[vbak["event_timestamp"] > min_extr_date]
        vbak = vbak[["VBELN", "event_timestamp"]].to_dict("r")
        vbak = {x["VBELN"]: x["event_timestamp"] for x in vbak}
    except:
        pass
    return vbak


def vbfa_closure(con, vbfa, min_extr_date):
    vbak = extract_vbak(con, min_extr_date=min_extr_date)
    vbeln_unique = set(vbfa["VBELN"].unique())
    vbfa_vbelv_types = list((x["VBELV"], x["VBTYP_V"]) for x in vbfa[["VBELV", "VBTYP_V"]].to_dict("records") if
                            x["VBELV"] not in vbeln_unique)
    # default the timestamp
    closure_events = {"VBELN": [x[0] for x in vbfa_vbelv_types], "VBTYP_N": [x[1] for x in vbfa_vbelv_types],
                      "event_timestamp": [vbak[x[0]] if x[0] in vbak else datetime.fromtimestamp(10000000) for x in vbfa_vbelv_types]}
    closure_df = pd.DataFrame(closure_events)
    return pd.concat([vbfa, closure_df]).sort_values("event_timestamp")


def apply(con, keep_first=True, min_extr_date="2020-01-01 00:00:00"):
    vbfa = con.prepare_and_execute_query("VBFA", ["ERDAT", "ERZET", "VBELN", "VBELV", "VBTYP_N", "VBTYP_V"])
    timestamp_column_from_dt_tm.apply(vbfa, "ERDAT", "ERZET", "event_timestamp")
    min_extr_date = parser.parse(min_extr_date)
    vbfa = vbfa[vbfa["event_timestamp"] > min_extr_date]
    doc_types = set(vbfa["VBTYP_N"].unique()).union(set(vbfa["VBTYP_V"].unique()))
    vbtyp = extract_vbtyp.apply_static(con, doc_types=doc_types)
    vbfa["VBTYP_N"] = vbfa["VBTYP_N"].map(vbtyp)
    vbfa["VBTYP_V"] = vbfa["VBTYP_V"].map(vbtyp)
    vbfa = vbfa_closure(con, vbfa, min_extr_date)
    vbfa["event_id"] = vbfa.index.astype(str)
    cols = {}
    for x in vbfa.columns:
        if x.startswith("event_"):
            cols[x] = x
        else:
            cols[x] = "event_" + x
    vbfa = vbfa.rename(columns=cols)
    vbfa["event_activity"] = "Create " + vbfa["event_VBTYP_N"]
    if not keep_first:
        vbfa["event_activity"] = vbfa["event_activity"] + " Item"
    vbfa["INVOLVED_DOCUMENTS"] = vbfa["event_VBELV"].astype(str) + constants.DOC_SEP + vbfa["event_VBELN"].astype(str)
    vbfa["INVOLVED_DOCUMENTS"] = vbfa["INVOLVED_DOCUMENTS"].apply(constants.set_documents)
    vbfa = vbfa.reset_index()
    return vbfa



from plantuml import *
import tempfile
import os


def apply(filename):
    with tempfile.TemporaryDirectory() as dir:
        dir_name = dir
    os.mkdir(dir_name)
    basename = os.path.basename(filename).split(".")[0] + ".png"
    pl = PlantUML(url="http://www.plantuml.com/plantuml/img/")
    pl.processes_file(filename, directory=dir_name, outfile=basename)
    fullpath = os.path.join(dir_name, basename)
    return fullpath



from sapextractor.algo.ap_ar import single_doc_transactions_dataframe, single_doc_transactions_log
from sapextractor.algo.ap_ar import document_flow_dataframe, document_flow_log
from sapextractor.algo.ap_ar import doc_flow_transactions_dataframe, doc_flow_transactions_log
from sapextractor.algo.ap_ar import obj_centr_log


def apply(con, ext_type, ext_arg):
    if ext_type == "single_doc_transactions_dataframe":
        return single_doc_transactions_dataframe.apply(con, **ext_arg)
    elif ext_type == "single_doc_transactions_log":
        return single_doc_transactions_log.apply(con, **ext_arg)
    elif ext_type == "document_flow_dataframe":
        return document_flow_dataframe.apply(con, **ext_arg)
    elif ext_type == "document_flow_log":
        return document_flow_log.apply(con, **ext_arg)
    elif ext_type == "doc_flow_transactions_dataframe":
        return doc_flow_transactions_dataframe.apply(con, **ext_arg)
    elif ext_type == "doc_flow_transactions_xes":
        return doc_flow_transactions_log.apply(con, **ext_arg)
    elif ext_type == "apar_obj_centr_log":
        return obj_centr_log.apply(con, **ext_arg)



from sapextractor.database_connection import oracle, sqlite
from sapextractor.algo.o2c import cli as o2c_cli
from sapextractor.algo.ap_ar import cli as ap_ar_cli
from sapextractor.algo.p2p import cli as p2p_cli


def main():
    print("== Connection to the database ==")
    print("Please pick one of the available connectors:")
    print("1) SQLite")
    print("2) Oracle")
    print()
    connector = input("Please insert your choice (default: 1) -> ")
    if not connector:
        connector = "1"
    con = None
    if connector == "1":
        con = sqlite.cli()
    elif connector == "2":
        con = oracle.cli()
    print()
    print("Please pick one of the supported processes:")
    print("a) O2C")
    print("b) Accounting (AP/AR)")
    print("c) P2P")
    print()
    process = input("Please insert your choice (default: a) -> ")
    if not process:
        process = "a"
    if process == "a":
        return o2c_cli.cli(con)
    elif process == "b":
        return ap_ar_cli.cli(con)
    elif process == "c":
        return p2p_cli.cli(con)



from sapextractor import algo, database_connection, utils, main, diagrams

__version__ = '0.0.13'
__doc__ = "SAP Extractor"
__author__ = 'PADS'
__author_email__ = 'alessandro.berti89@gmail.com'
__maintainer__ = 'PADS'
__maintainer_email__ = 'alessandro.berti89@gmail.com'


def connect_sqlite(path='./sap.sqlite'):
    return database_connection.sqlite.apply(path=path)


def connect_oracle(hostname="127.0.0.1", port="1521", sid="xe", username="system", password="oracle"):
    return database_connection.oracle.apply(hostname=hostname, port=port, sid=sid, username=username, password=password)


def get_o2c_classic_event_log(con, ref_type="Invoice", keep_first=True, min_extr_date="2020-01-01 00:00:00", gjahr="2020", enable_changes=True, enable_payments=True):
    return algo.o2c.o2c_1d_log_extractor.apply(con, ref_type=ref_type, keep_first=keep_first,
                                               min_extr_date=min_extr_date, gjahr=gjahr, enable_changes=enable_changes, enable_payments=enable_payments)


def get_o2c_classic_dataframe(con, ref_type="Invoice", keep_first=True, min_extr_date="2020-01-01 00:00:00", gjahr="2020", enable_changes=True, enable_payments=True):
    return algo.o2c.o2c_1d_dataframe_extractor.apply(con, ref_type=ref_type, keep_first=keep_first,
                                                     min_extr_date=min_extr_date, gjahr=gjahr, enable_changes=enable_changes, enable_payments=enable_payments)


def get_o2c_obj_centr_log(con, keep_first=True, min_extr_date="2020-01-01 00:00:00", gjahr="2020", enable_changes=True, enable_payments=True):
    return algo.o2c.obj_centr_log.apply(con, keep_first=keep_first, min_extr_date=min_extr_date, gjahr=gjahr, enable_changes=enable_changes, enable_payments=enable_payments)


def get_p2p_classic_event_log(con, ref_type="EKKO"):
    return algo.p2p.p2p_1d_log.apply(con, ref_type=ref_type)


def get_p2p_classic_dataframe(con, ref_type="EKKO"):
    return algo.p2p.p2p_1d_dataframe.apply(con, ref_type=ref_type)


def get_p2p_obj_centr_log(con):
    return algo.p2p.obj_centr_log.apply(con)


def get_ap_ar_single_doc_transactions_dataframe(con, gjahr="1997", bukrs=None):
    return algo.ap_ar.single_doc_transactions_dataframe.apply(con, gjahr=gjahr, bukrs=bukrs)


def get_ap_ar_single_doc_transactions_log(con, gjahr="1997", bukrs=None):
    return algo.ap_ar.single_doc_transactions_log.apply(con, gjahr=gjahr, bukrs=bukrs)


def get_ap_ar_document_flow_dataframe(con, ref_type="Goods receipt", gjahr="1997", bukrs=None):
    return algo.ap_ar.document_flow_dataframe.apply(con, ref_type=ref_type, gjahr=gjahr, bukrs=bukrs)


def get_ap_ar_document_flow_log(con, ref_type="Goods receipt", gjahr="1997", bukrs=None):
    return algo.ap_ar.document_flow_log.apply(con, ref_type=ref_type, gjahr=gjahr, bukrs=bukrs)


def get_ap_ar_doc_flow_transactions_dataframe(con, ref_type="Goods receipt", gjahr="1997", bukrs=None):
    return algo.ap_ar.doc_flow_transactions_dataframe.apply(con, ref_type=ref_type, gjahr=gjahr, bukrs=bukrs)


def get_ap_ar_doc_flow_transactions_log(con, ref_type="Goods receipt", gjahr="1997", bukrs=None):
    return algo.ap_ar.doc_flow_transactions_log.apply(con, ref_type=ref_type, gjahr=gjahr, bukrs=bukrs)


def get_ap_ar_obj_centr_log(con, gjahr="1997", bukrs=None):
    return algo.ap_ar.obj_centr_log.apply(con, gjahr=gjahr, bukrs=bukrs)


def cli():
    return main.main()



def apply(con, target_language="E"):
    dict_field_desc = {}
    try:
        df = con.prepare_and_execute_query("DD03T", ["DDLANGUAGE", "FIELDNAME", "DDTEXT"], " WHERE DDLANGUAGE = '"+target_language+"'")
        stream = df.to_dict("records")
        for el in stream:
            dict_field_desc[el["FIELDNAME"]] = el["DDTEXT"]
    except:
        pass
    return dict_field_desc



import pandas as pd
from datetime import datetime
from sapextractor.utils import constants


def apply(dataframe, dt_column, tm_column, target_column):
    try:
        dataframe[target_column] = dataframe[dt_column] + " " + dataframe[tm_column]
        dataframe[target_column] = pd.to_datetime(dataframe[target_column], format=constants.TIMESTAMP_FORMAT)
        dataframe = dataframe.sort_values("event_timestamp")
        dataframe = dataframe.dropna(subset=["event_timestamp"], how="any")
    except:
        dataframe[dt_column] = pd.to_datetime(dataframe[dt_column])
        dataframe[tm_column] = pd.to_datetime(dataframe[tm_column])
        dataframe[dt_column] = dataframe[dt_column].apply(lambda x: x.timestamp())
        dataframe[tm_column] = dataframe[tm_column].apply(lambda x: x.timestamp())

        if len(dataframe) > 0:
            dataframe[target_column] = dataframe[dt_column] + dataframe[tm_column]
            dataframe[target_column] = dataframe[target_column].apply(lambda x: datetime.fromtimestamp(x))
            dataframe = dataframe.sort_values("event_timestamp")
            dataframe = dataframe.dropna(subset=["event_timestamp"], how="any")

    return dataframe



from sapextractor.algo.p2p import p2p_1d_dataframe, p2p_1d_log


def apply(con, ext_type, ext_arg):
    if ext_type == "p2p_1d_dataframe":
        return p2p_1d_dataframe.apply(con, **ext_arg)
    elif ext_type == "p2p_1d_log":
        return p2p_1d_log.apply(con, **ext_arg)



import pandas as pd
from pm4pymdl.objects.mdl.exporter import exporter as mdl_exporter
from pm4pymdl.objects.ocel.exporter import exporter as ocel_exporter

from sapextractor.algo.o2c import o2c_common
from sapextractor.utils import constants
from sapextractor.utils.change_tables import extract_change


def get_changes(con, dataframe):
    vbeln_values = dataframe["event_VBELN"].unique()
    ret = []
    for tup in [("VERKBELEG", "VBAK"), ("VERKBELEG", "VBAP"), ("VERKBELEG", "VBUK"), ("LIEFERUNG", "LIKP"),
                ("LIEFERUNG", "LIPS"), ("LIEFERUNG", "VBUK")]:
        changes = extract_change.apply(con, objectclas=tup[0], tabname=tup[1])
        changes = {x: y for x, y in changes.items() if x in vbeln_values}
        for x, y in changes.items():
            y["event_VBELN"] = y["event_AWKEY"]
            y["event_activity"] = y["event_CHANGEDESC"]
            y["INVOLVED_DOCUMENTS"] = y["event_AWKEY"]
            y["INVOLVED_DOCUMENTS"] = y["INVOLVED_DOCUMENTS"].apply(constants.set_documents)
            ret.append(y)

    if ret:
        ret = pd.concat(ret)
    else:
        ret = pd.DataFrame()

    return ret


def apply(con, keep_first=True, min_extr_date="2020-01-01 00:00:00", gjahr="2020", enable_changes=True,
          enable_payments=True):
    dataframe = o2c_common.apply(con, keep_first=keep_first, min_extr_date=min_extr_date)
    dataframe = dataframe.sort_values("event_timestamp")
    if enable_changes:
        changes = get_changes(con, dataframe)
    else:
        changes = pd.DataFrame()
    if len(changes) > 0:
        changes = changes.sort_values("event_timestamp")
    dataframe = pd.concat([dataframe, changes])
    dataframe["event_id"] = dataframe.index.astype(str)
    dataframe = dataframe.sort_values("event_timestamp")
    dataframe.type = "succint"
    return dataframe


def cli(con):
    print("\n\nO2C Object-Centric Log Extractor\n\n")
    min_extr_date = input("Insert the minimum extraction date (default: 2020-01-01 00:00:00): ")
    if not min_extr_date:
        min_extr_date = "2020-01-01 00:00:00"
    gjahr = input("Insert the fiscal year (default: 2020):")
    if not gjahr:
        gjahr = "2020"
    dataframe = apply(con, min_extr_date=min_extr_date, gjahr=gjahr)
    path = input("Insert the path where the log should be saved (default: o2c.xmlocel): ")
    if not path:
        path = "o2c.xmlocel"
    if path.endswith("mdl"):
        mdl_exporter.apply(dataframe, path)
    elif path.endswith("jsonocel") or path.endswith("xmlocel"):
        ocel_exporter.apply(dataframe, path)



from sapextractor.algo.ap_ar import single_doc_transactions_dataframe, single_doc_transactions_log
from sapextractor.algo.ap_ar import document_flow_dataframe, document_flow_log
from sapextractor.algo.ap_ar import doc_flow_transactions_dataframe, doc_flow_transactions_log
from sapextractor.algo.ap_ar import obj_centr_log


def cli(con):
    print("\n\nAccounting (AP/AR) extraction\n")
    print("available extraction types:")
    print("1) Transactions for the single document (XES log)")
    print("2) Transactions for the single document (dataframe)")
    print("3) Document Flow (XES log)")
    print("4) Document Flow (dataframe)")
    print("5) Document Flow Transactions (XES log)")
    print("6) Document Flow Transactions (dataframe)")
    print("7) Accounting Object-Centric Event Log")
    print()
    ext_type = input("insert your choice (default: 1):")
    if not ext_type:
        ext_type = "1"
    if ext_type == "1":
        return single_doc_transactions_log.cli(con)
    elif ext_type == "2":
        return single_doc_transactions_dataframe.cli(con)
    elif ext_type == "3":
        return document_flow_log.cli(con)
    elif ext_type == "4":
        return document_flow_dataframe.cli(con)
    elif ext_type == "5":
        return doc_flow_transactions_log.cli(con)
    elif ext_type == "6":
        return doc_flow_transactions_dataframe.cli(con)
    elif ext_type == "7":
        return obj_centr_log.cli(con)



from sapextractor.algo.ap_ar import ap_ar_common
from pm4pymdl.objects.ocel.exporter import exporter as jmd_exporter
from pm4pymdl.objects.mdl.exporter import exporter as mdl_exporter


def apply(con, gjahr="1997", bukrs=None, **ext_arg):
    bkpf = ap_ar_common.get_full_dataframe(con, gjahr=gjahr, bukrs=bukrs, filter_columns=False)
    bkpf["event_activity"] = bkpf["event_ONLYACT"] + " (" + bkpf["event_BLART"] + ")"
    bkpf["event_id"] = bkpf.index.astype(str)
    bkpf = bkpf.sort_values("event_timestamp")
    bkpf.type = "succint"
    return bkpf


def cli(con):
    print("\n\nAccounting Object-Centric Log Extractor\n\n")
    dataframe = apply(con)
    path = input("Insert the path where the log should be saved (default: accounting.mdl): ")
    if not path:
        path = "accounting.xmlocel"
    if path.endswith("mdl"):
        mdl_exporter.apply(dataframe, path)
    elif path.endswith("jsonocel") or path.endswith("xmlocel"):
        jmd_exporter.apply(dataframe, path)




def ekko_mkpf_connection(con):
    mseg = con.prepare_and_execute_query("MSEG", ["MBLNR", "EBELN"])
    mseg = mseg.dropna(subset=["MBLNR", "EBELN"], how="any")
    mseg = mseg.to_dict('records')
    mseg = [("EKKO_"+x["EBELN"], "MKPF_"+x["MBLNR"]) for x in mseg]
    return mseg



from datetime import datetime

import pandas as pd

from sapextractor.utils import constants
from sapextractor.utils.blart import extract_blart
from sapextractor.utils.change_tables import extract_change
from sapextractor.utils.dates import timestamp_column_from_dt_tm
from sapextractor.utils.tstct import extract_tstct


def extract_bkpf(con, gjahr="1997", bukrs=None):
    additional_query_part = ""
    if gjahr is not None and bukrs is not None:
        additional_query_part = " WHERE GJAHR = '"+gjahr+"' AND BUKRS = '"+bukrs+"'"
    elif gjahr is not None:
        additional_query_part = " WHERE GJAHR = '" + gjahr + "'"
    elif bukrs is not None:
        additional_query_part = " WHERE BUKRS = '" + bukrs + "'"
    bkpf = con.prepare_and_execute_query("BKPF", ["BELNR", "BLART", "CPUDT", "CPUTM", "USNAM", "TCODE", "AWKEY"], additional_query_part=additional_query_part)
    bkpf = bkpf.dropna(subset=["BELNR", "TCODE", "BLART"], how="any")
    transactions = set(bkpf["TCODE"].unique())
    doc_types = set(bkpf["BLART"].unique())
    tcodes = extract_tstct.apply_static(con, transactions=transactions)
    blart = extract_blart.apply_static(con, doc_types=doc_types)
    cols = {x: "event_" + x for x in bkpf.columns}
    bkpf = bkpf.rename(columns=cols)
    bkpf["event_ONLYACT"] = bkpf["event_TCODE"].map(tcodes)
    bkpf["event_BLART"] = bkpf["event_BLART"].map(blart)
    bkpf = bkpf.dropna(subset=["event_ONLYACT", "event_BLART"], how="any")
    bkpf["INVOLVED_DOCUMENTS"] = bkpf["event_BELNR"].astype(str)
    bkpf["INVOLVED_DOCUMENTS"] = bkpf["INVOLVED_DOCUMENTS"].apply(constants.set_documents)
    bkpf = timestamp_column_from_dt_tm.apply(bkpf, "event_CPUDT", "event_CPUTM", "event_timestamp")
    bkpf_first = bkpf.groupby("event_BELNR").first().reset_index()
    first_stream = bkpf_first[["event_BELNR", "event_timestamp", "event_BLART"]].to_dict("records")
    doc_first_dates = {x["event_BELNR"]: x["event_timestamp"] for x in first_stream}
    doc_types = {x["event_BELNR"]: x["event_BLART"] for x in first_stream}

    return bkpf, doc_first_dates, doc_types


def extract_bseg(con, doc_first_dates, doc_types, gjahr="1997", bukrs=None):
    additional_query_part = ""
    if gjahr is not None and bukrs is not None:
        additional_query_part = " WHERE GJAHR = '"+gjahr+"' AND BUKRS = '"+bukrs+"'"
    elif gjahr is not None:
        additional_query_part = " WHERE GJAHR = '" + gjahr + "'"
    elif bukrs is not None:
        additional_query_part = " WHERE BUKRS = '" + bukrs + "'"
    bseg = con.prepare_and_execute_query("BSEG", ["BELNR", "GJAHR", "BUZEI", "AUGDT", "AUGBL"], additional_query_part=additional_query_part)
    bseg = bseg.dropna(subset=["BELNR", "AUGBL", "AUGDT"], how="any")
    bseg["BELNR_TYPE"] = bseg["BELNR"].map(doc_types)
    bseg["AUGBL_TYPE"] = bseg["AUGBL"].map(doc_types)
    bseg = bseg.dropna(subset=["BELNR_TYPE", "AUGBL_TYPE"])
    bseg["AUGDT"] = pd.to_datetime(bseg["AUGDT"])
    bseg["AUGDT"] = bseg["AUGDT"].apply(lambda x: x.timestamp())
    bseg["AUGDT"] += 86399
    bseg["AUGDT"] = bseg["AUGDT"].apply(lambda x: datetime.fromtimestamp(x))
    cols = {x: "event_" + x for x in bseg.columns}
    bseg = bseg.rename(columns=cols)
    bseg["INVOLVED_DOCUMENTS"] = bseg["event_BELNR"].astype(str) + constants.DOC_SEP + bseg["event_AUGBL"].astype(str)
    bseg["INVOLVED_DOCUMENTS"] = bseg["INVOLVED_DOCUMENTS"].apply(constants.set_documents)
    bseg["event_timestamp"] = bseg["event_AUGDT"]
    bseg = bseg.dropna(subset=["event_timestamp"], how="any")
    bseg["event_ONLYACT"] = "Clear Document"
    bseg["event_DOCTYPE"] = bseg["event_BELNR"].map(doc_types)

    return bseg


def extract_changes_bkpf(con, bkpf, doc_types, gjahr="1997", bukrs=None):
    changes = extract_change.get_changes_dataframe_two_mapping(con, bkpf, "event_AWKEY", "event_BELNR",
                                                               resource_column="event_USNAM", objectclas="BELEG",
                                                               tabname="BKPF")
    if len(changes) > 0:
        changes["event_BLART"] = changes["event_BELNR"].map(doc_types)
        changes = changes.dropna(subset=["event_BLART"], how="any")
        changes["INVOLVED_DOCUMENTS"] = changes["event_BELNR"].astype(str)
        changes["INVOLVED_DOCUMENTS"] = changes["INVOLVED_DOCUMENTS"].apply(constants.set_documents)
    return changes


def get_single_dataframes(con, gjahr="1997", bukrs=None, filter_columns=False):
    bkpf, doc_first_dates, doc_types = extract_bkpf(con, gjahr=gjahr, bukrs=bukrs)
    #bkpf["@@DOCTYPE"] = bkpf["event_BELNR"].map(doc_types)
    #bkpf.to_csv("bkpf.csv", index=False)
    bseg = extract_bseg(con, doc_first_dates, doc_types, gjahr=gjahr, bukrs=bukrs)
    #bseg["@@DOCTYPE"] = bseg["event_BELNR"].map(doc_types)
    #bseg.to_csv("bseg.csv", index=False)
    #changes = extract_changes_bkpf(con, bkpf, doc_types, gjahr=gjahr, bukrs=bukrs)
    if filter_columns:
        bkpf = bkpf[[x for x in bkpf.columns if x.startswith("event_")]]
        bseg = bseg[[x for x in bseg.columns if x.startswith("event_")]]
    #bkpf = pd.concat([bkpf, changes])
    return bkpf, bseg


def get_full_dataframe(con, gjahr="1997", bukrs=None, filter_columns=False):
    bkpf, bseg = get_single_dataframes(con, gjahr=gjahr, bukrs=bukrs, filter_columns=filter_columns)
    return pd.concat([bkpf, bseg])
